/* $Id: ff.h,v 3.13 1998/02/16 12:12:25 kas Exp $ */

/* Definice struktury force_field, ktera popisuje silove pole molekuly */
/* Tato struktura muze byt rozsirena jeste o dalsi cleny */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995-1997  Jan "Yenya" Kasprzak <kas@fi.muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef FF_H__
#define FF_H__	1

#include "proto.h"

/* For now, the local minimization routine requires REAL == double */
#ifndef REAL
#define REAL double
#endif

#ifndef USE_TRANSFORMED
#define NONTRANSFORMED_ONLY
#endif

#ifndef NONTRANSFORMED_ONLY
#include "trans.h"
#endif

#define ATOM_NAME_LENGTH	4

/* This is a Opal's equivalent of AMBER's parmstruct */
typedef struct force_field_
{
	char *title;	/* Amber struct parm.ititl */
	int natoms, natypes, nhydronb, nbonds, nbondtypes, nangles,
		nangletypes, ndihs, ndihtypes, nresidues;
	char **anames,	/* Has [natoms][ATOM_NAME_LENGTH] elements */
		**atypenames;	/* [natoms][ATOM_NAME_LENGTH] */
			/* Nonbonded interaction parameters */
	REAL diel_lin, diel_const;	/* for nonbonded interactions */
	REAL cutoff_sq;			/* 1991 force field cutoff
					 * (square of this value) */
	REAL vdw_14, eel_14;		/* for 1-4 interactions */
	REAL *charge;	/* [natoms] */
	REAL *r6, *r12;	/* [natypes*(natypes+1)/2] */
	REAL *h10, *h12;	/* [nhydronb] */
	REAL *rk, *req;		/* Bond parameters [nbondtypes] */
	REAL *tk, *teq;		/* Angle parameters [nangletypes] */
	REAL *pk;		/* Dihedral parameters [ndihtypes] */
	int *pn, *phase;	/* Dihedral parameters [ndihtypes]
				 * pn is an integer in range -4 .. 4,
				 * phase is 1 iff delta==0 and -1
				 * iff delta == +/-pi
				 */
	int *atypes;		/* [natoms] */
	int *residue;	/* [natoms] */
	int *nbnum;		/* [natypes*natypes] */
	int *bondat1, *bondat2, *bondnum;	/* [nbonds] */
	int *angleat1, *angleat2, *angleat3, *anglenum;	/* [nangles] */
	int *dihat1, *dihat2, *dihat3, *dihat4, *dihnum;	/* [ndihs] */
	int *improper;		/* Boolean: Improper dihedral? [ndihs] */
	int *ignored;		/* Boolean: This dihedral is to be
				 * ignored in the force field [ndihs]
				 */
#ifndef NONTRANSFORMED_ONLY
	REAL lambda;

	pair_int *pip;
	angle_int *aip;
	dih_int *dip;

	int *tanglenum;
	int *tdihnum;
#endif
} force_field;

extern force_field *read_ff __P((char *fname));
extern force_field *write_ff __P((char *fname));
extern int check_ff __P((force_field *ffp));
extern void malloc_ff __P((force_field *ffp));
extern void free_ff __P((force_field *ffp));
extern void setup_ff __P((force_field *ffp, REAL vdw14, REAL eel14,
	REAL diel_lin, REAL diel_const, REAL cutoff));

extern int find_bond __P((force_field *ffp, int a1, int a2));
extern int find_angle __P((force_field *ffp, int a1, int a2, int a3));

extern force_field *prm2ff __P((char *filename));

REAL compute_bond __P((force_field *ffp, REAL *x, REAL *y, REAL *z, int bond));
REAL compute_angle __P((force_field *ffp, REAL *x, REAL *y, REAL *z, int angle));
REAL compute_dih __P((force_field *ffp, REAL *x, REAL *y, REAL *z, int dih));
REAL compute_14 __P((force_field *ffp, REAL *x, REAL *y, REAL *z,
	int a1, int a2));
REAL compute_nb __P((force_field *ffp, REAL *x, REAL *y, REAL *z,
	int a1, int a2));

void diff_bond __P((force_field *ffp, REAL *x, REAL *y, REAL *z, REAL *dx,
	REAL *dy, REAL *dz, int bond));
void diff_angle __P((force_field *ffp, REAL *x, REAL *y, REAL *z, REAL *dx,
	REAL *dy, REAL *dz, int angle));
void diff_dih __P((force_field *ffp, REAL *x, REAL *y, REAL *z, REAL *dx,
	REAL *dy, REAL *dz, int dih));
void diff_14 __P((force_field *ffp, REAL *x, REAL *y, REAL *z, REAL *dx,
	REAL *dy, REAL *dz, int a1, int a2));
void diff_nb __P((force_field *ffp, REAL *x, REAL *y, REAL *z, REAL *dx,
	REAL *dy, REAL *dz, int a1, int a2));
#ifndef NONTRANSFORMED_ONLY

extern int read_trans_ff __P((force_field *ffp, REAL lambda,
	char *atmpl, char *ttmpl, char *ptmpl));
extern void free_trans_part __P((force_field *ffp));
#endif /* NONTRANSFORMED_ONLY */

#endif /* FF_H__ */
