/* $Id: trans.c,v 3.2 1997/03/10 00:03:02 kas Exp $ */

/* Rutiny pro vypocet slozek transformovaneho siloveho pole */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995-1997  Jan "Yenya" Kasprzak <kas@fi.muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <math.h>

#define USE_TRANSFORMED	1
#include "ff.h"

/* #define PRINT_DEBUG 1 */

REAL compute_bond(ffp, x, y, z, bond)
force_field *ffp;
REAL *x, *y, *z;
int bond;
{
	REAL b, result;
	int at1, at2;

	at1 = ffp->bondat1[bond];
	at2 = ffp->bondat2[bond];

#ifdef PRINT_DEBUG
	printf("%4d-%4d %6.3f %6.3f ",
		at1+1, at2+1, ffp->rk[ffp->bondnum[bond]],
		ffp->req[ffp->bondnum[bond]]);
#endif
	b = sqrt((x[at1] - x[at2])*(x[at1] - x[at2])
		+ (y[at1] - y[at2])*(y[at1] - y[at2])
		+ (z[at1] - z[at2])*(z[at1] - z[at2]));
#ifdef PRINT_DEBUG
	printf("%6.3f ", b);
#endif
#ifndef TRANSFORMED_BONDS_BY_TABLES
	b -= ffp->req[ffp->bondnum[bond]];
	result = b*b;
/*	result += 3*ffp->lambda*ffp->lambda; */
#else
	idx = b/ffp->pip->step;
	if (idx >= ffp->pip->dim-1)
		idx = ffp->pip->dim-2;
	dx = (b-idx*ffp->pip->step)/ffp->pip->step;
/*	printf ("idx=%d, dx=%f, ",idx,dx);  */
	/* b^2 */
	result = ffp->pip->val_2[idx]
		+ dx * (ffp->pip->val_2[idx+1] - ffp->pip->val_2[idx]);
#ifdef PRINT_DEBUG
	printf ("b=%f,",result);
#endif
	/* -2*b*b0 */
#if 0
	result += foo = -2*ffp->req[ffp->bondnum[bond]] *
		(ffp->pip->val_1[idx] + dx * (ffp->pip->val_1[idx+1]
			- ffp->pip->val_1[idx]));
#else
	result += foo = -2*b*ffp->req[ffp->bondnum[bond]];
#endif
#ifdef PRINT_DEBUG
	printf ("%f,",foo);
#endif
	/* + b0^2 */
	result += foo = ffp->req[ffp->bondnum[bond]]*ffp->req[ffp->bondnum[bond]];
#ifdef PRINT_DEBUG
	printf ("%f ",foo);
#endif
#endif /* TRANSFORMED_BONDS */
	result *= ffp->rk[ffp->bondnum[bond]];
#ifdef PRINT_DEBUG
	printf ("%3.3f\n", result);
#endif
	return result;
}

#define EPSILON 1e-4

REAL compute_angle(ffp, x, y, z, angle)
force_field *ffp;
REAL *x, *y, *z;
int angle;
{
	REAL a;
	int at1, at2, at3;
	int tnum, dim, idx;
	REAL dx;

	at1 = ffp->angleat1[angle];
	at2 = ffp->angleat2[angle];
	at3 = ffp->angleat3[angle];

#ifdef PRINT_DEBUG
	printf ("%3d-%3d-%3d %3.3f %3.3f ",
		at1+1, at2+1, at3+1, ffp->tk[ffp->anglenum[angle]],
		ffp->teq[ffp->anglenum[angle]]*180.0/M_PI);
#endif
	if (x[at1] - x[at2] == 0.0 && y[at1] - y[at2] == 0.0
		&& z[at1] - z[at2] == 0.0)
	{
		a = -1.0;
		goto computed;
	}

	if (x[at3] - x[at2] == 0.0 && y[at3] - y[at2] == 0.0
		&& z[at3] - z[at2] == 0.0)
	{
		a = -1.0;
		goto computed;
	}

	a = (x[at1] - x[at2])*(x[at3] - x[at2])
		+ (y[at1] - y[at2])*(y[at3] - y[at2])
		+ (z[at1] - z[at2])*(z[at3] - z[at2]);

	a /= sqrt(((x[at1] - x[at2])*(x[at1] - x[at2])
		+ (y[at1] - y[at2])*(y[at1] - y[at2])
		+ (z[at1] - z[at2])*(z[at1] - z[at2]))
		* ((x[at3] - x[at2])*(x[at3] - x[at2])
		+ (y[at3] - y[at2])*(y[at3] - y[at2])
		+ (z[at3] - z[at2])*(z[at3] - z[at2])));


	/* FIXME: Tady se mozna bude muset zohlednit, jestli ten
	 * rozdil nepreleze ven z intervalu [-pi,pi].
	 */
computed:
	a = acos(a);
#ifdef PRINT_DEBUG
	printf ("%3.3f ", a*180.0/M_PI);
#endif
	tnum = ffp->tanglenum[angle];
	dim = ffp->aip[tnum].dim - 1;
	if (fabs(a - M_PI) < EPSILON)
		a = ffp->aip[tnum].val[dim];
	else {
		idx = a*dim/M_PI;
		dx = (a-idx*M_PI/dim)*dim/M_PI;
		a = ffp->aip[tnum].val[idx] + dx *
			(ffp->aip[tnum].val[idx+1] - ffp->aip[tnum].val[idx]);
#ifdef PRINT_DEBUG
		printf ("idx=%d, dx=%f, a=%f, ",idx,dx,a);
#endif
	}
	a *= ffp->tk[ffp->anglenum[angle]];
#ifdef PRINT_DEBUG
	printf ("%3.3f\n", a);
#endif
	return a;
}

REAL compute_dih(ffp, x, y, z, dih)
force_field *ffp;
REAL *x, *y, *z;
int dih;
{
	int a1, a2, a3, a4;
	REAL a[3], b[3], c[3];
	REAL result = 0.0, tmp, n1,n2, ang1, ang2, dx;
	int dihnum, i1, i2, idx, tnum, dim1, dim2, dim;

	a1 = ffp->dihat1[dih];
	a2 = ffp->dihat2[dih];
	a3 = ffp->dihat3[dih];
	a4 = ffp->dihat4[dih];
	dihnum = ffp->dihnum[dih];
#ifdef PRINT_DEBUG
	printf ("%4d-%4d-%4d-%4d (%4d) %3.3f %3.3f ", a1+1, a2+1, a3+1, a4+1,
		dihnum,
		ffp->phase[dihnum] == 1 ? 0.0 : 180.0, ffp->pk[dihnum]);
#endif
	if (ffp->pk[dihnum] <= 0.0)
		goto zero_peak;
	/* Normalovy vektor k rovine 1-2-3 */
	a[0] = x[a1] - x[a2];
	a[1] = y[a1] - y[a2];
	a[2] = z[a1] - z[a2];
	b[0] = x[a3] - x[a2];
	b[1] = y[a3] - y[a2];
	b[2] = z[a3] - z[a2];
	ang1 = a[0]*b[0]+a[1]*b[1]+a[2]*b[2];
	ang1 /= sqrt((a[0]*a[0]+a[1]*a[1]+a[2]*a[2])
		*(b[0]*b[0]+b[1]*b[1]+b[2]*b[2]));
	ang1 = acos(ang1);
	c[2] = a[0]*b[1] - a[1]*b[0];	n1 =  c[2]*c[2];
	c[1] = a[2]*b[0] - a[0]*b[2];	n1 += c[1]*c[1];
	c[0] = a[1]*b[2] - a[2]*b[1];	n1 += c[0]*c[0];

	/* Normalovy vektor k rovine 2-3-4 */
	a[0] = x[a4] - x[a3];
	a[1] = y[a4] - y[a3];
	a[2] = z[a4] - z[a3];
	ang2 = -a[0]*b[0] -a[1]*b[1] -a[2]*b[2];
	ang2 /= sqrt((a[0]*a[0]+a[1]*a[1]+a[2]*a[2])
		*(b[0]*b[0]+b[1]*b[1]+b[2]*b[2]));
	ang2 = acos(ang2);
	tmp = b[1]*a[0] - b[0]*a[1]; n2  = tmp*tmp; result  = c[2]*tmp;
	tmp = b[0]*a[2] - b[2]*a[0]; n2 += tmp*tmp; result += c[1]*tmp;
	tmp = b[2]*a[1] - b[1]*a[2]; n2 += tmp*tmp; result += c[0]*tmp;

	if (!(n1*=n2))
		return -1.0;

	result /= sqrt(n1);
	result = acos(result);

	/* Nyni mame v result <(1-2-3,2-3-4).
	 * V ang1 je <(1-2-3) a v ang2 je <(2-3-4).
	 */
	tnum = ffp->tdihnum[dih];
	dim1 = ffp->dip[tnum].adim1-1;
	dim2 = ffp->dip[tnum].adim2-1;
	i1 = rint(ang1/M_PI*dim1);
	i2 = rint(ang2/M_PI*dim2);
	dim = ffp->dip[tnum].tdim-1;
	idx = result/M_PI*dim;
	if (idx >= dim)
		idx--;
#if 0
	printf("<%d %d %d>", idx, i1, i2);
#endif
#ifdef PRINT_DEBUG
	printf(" %f ", result*180/M_PI);
/*	printf(" %f %f %f ", result*180/M_PI, ang1*180/M_PI, ang2*180/M_PI); */
#endif
	dx = (result - idx*M_PI/dim)*dim/M_PI;
	idx = (dim1+1)*(dim2+1)*idx + (dim1+1)*i1 + i2;
	result = ffp->dip[tnum].val[idx] + dx *
		(ffp->dip[tnum].val[idx+(dim1+1)*(dim2+1)]
		- ffp->dip[tnum].val[idx]);
#if 0
	printf (" %f %d %f %d ", dx, idx, result, tnum);
#endif
	result = ffp->pk[dihnum]*(1 + result);
zero_peak:
#ifdef PRINT_DEBUG
	printf ("%3.3f\n", result);
#endif
	if (ffp->ignored[dih])
		return result;
	compute_nb(ffp, x,y,z, a1,a4, &n1, &n2);
	return result + n1/ffp->vdw_14 + n2/ffp->eel_14;
}

REAL compute_nb(ffp, x,y,z, a1,a2, vdw, eel)
force_field *ffp;
REAL *x, *y, *z;
int a1, a2;
REAL *vdw, *eel;
{
	REAL r, dx;
	int i, idx;

	i = ffp->nbnum[ffp->natypes*ffp->atypes[a1]+ffp->atypes[a2]];
#if 0
	if (i < 0)
		printf("Hydrogen bond found between atoms %d and %d\n",
			a1+1, a2+1);
#endif
	r = (x[a1]-x[a2])*(x[a1]-x[a2])
		+(y[a1]-y[a2])*(y[a1]-y[a2])
		+(z[a1]-z[a2])*(z[a1]-z[a2]);

	if (r <= 0.0)
	{
		*vdw = 1E+20;
		*eel = 1E+20;
		return 2E+20;
	}

	r = sqrt(r);
	idx = r/ffp->pip->step;
	if (idx >= ffp->pip->dim-1)
		idx = ffp->pip->dim-2;
	dx = (r-idx*ffp->pip->step)/ffp->pip->step;
	if (i >= 0)
		*vdw = ffp->r12[i] * (ffp->pip->val_m12[idx] +
			dx * (ffp->pip->val_m12[idx+1]-ffp->pip->val_m12[idx]))
			- ffp->r6[i] * (ffp->pip->val_m6[idx] +
			dx * (ffp->pip->val_m6[idx+1]-ffp->pip->val_m6[idx]));
	else
		*vdw = ffp->h12[-i-2] * (ffp->pip->val_m12[idx] +
			dx * (ffp->pip->val_m12[idx+1]-ffp->pip->val_m12[idx]))
			- ffp->h10[-i-2] * (ffp->pip->val_m10[idx] +
			dx * (ffp->pip->val_m10[idx+1]-ffp->pip->val_m10[idx]));

	/* Muzeme udelat nasledujici, protoze diel_lin*diel_const = 0 */
	*eel = ffp->charge[a1]*ffp->charge[a2]
		* (ffp->diel_lin * (ffp->pip->val_m2[idx] +
		dx * (ffp->pip->val_m2[idx+1] - ffp->pip->val_m2[idx]))
		+ ffp->diel_const * (ffp->pip->val_m1[idx] + 
		dx * (ffp->pip->val_m1[idx+1] - ffp->pip->val_m1[idx])));
#ifdef PRINT_DEBUG
	printf("%3d-%3d Nonbonded: r=%3.3f, vdw=%3.3f%s, eel=%3.3f\n",
		a1+1,a2+1, r, *vdw,idx<0?" (hydrogen)":"", *eel);
#endif
	return *vdw + *eel;
}
